(******************************************************************************
 * Isabelle/C
 *
 * Copyright (c) 2018-2019 Université Paris-Saclay, Univ. Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)

chapter \<open>Isabelle/C\<close>

section \<open>Core Language: An Abstract Syntax Tree Definition (C Language without Annotations)\<close>

theory C_Ast
  imports Main
begin

subsection \<open>Loading the Generated AST\<close>

text \<open> The abstract syntax tree of the C language considered in the Isabelle/C project is
arbitrary, but it must already come with a grammar making the connection with a default AST, so that
both the grammar and AST can be imported to SML.\<^footnote>\<open>Additionally, the grammar and AST
must both have a free licence --- compatible with the Isabelle AFP, for them to be publishable
there.\<close> The Haskell Language.C project fulfills this property: see for instance
\<^url>\<open>http://hackage.haskell.org/package/language-c\<close> and
\<^url>\<open>https://github.com/visq/language-c/blob/master/src/Language/C/Syntax/AST.hs\<close>,
where its AST is being imported in the present theory file \<^file>\<open>C_Ast.thy\<close>, whereas
its grammar will be later in \<^file>\<open>C_Parser_Language.thy\<close>
(\<^file>\<open>C_Parser_Language.thy\<close> depends on \<^file>\<open>C_Ast.thy\<close>). The AST
importation is based on a modified version of Haskabelle, which generates the C AST from Haskell to
an ML file. \<close>

ML \<comment> \<open>\<^file>\<open>../generated/c_ast.ML\<close>\<close> \<open>
val fresh_ident0 =
  let val i = Synchronized.var "counter for new identifier" 38 in
    fn () => Int.toString (Synchronized.change_result i (fn i => (i, i + 1)))
  end
\<close>

ML \<comment> \<open>\<^file>\<open>../generated/c_ast.ML\<close>\<close> \<open>
\<comment> \<open>\<^file>\<open>../../Citadelle/src/compiler_generic/meta_isabelle/Printer_init.thy\<close>\<close>
structure CodeType = struct
  type mlInt = string
  type 'a mlMonad = 'a option
end

structure CodeConst = struct
  structure Monad = struct
    val bind = fn
        NONE => (fn _ => NONE)
      | SOME a => fn f => f a
    val return = SOME
  end

  structure Printf = struct
    local
      fun sprintf s l =
        case String.fields (fn #"%" => true | _ => false) s of
          [] => ""
        | [x] => x
        | x :: xs =>
            let fun aux acc l_pat l_s =
              case l_pat of
                [] => rev acc
              | x :: xs => aux (String.extract (x, 1, NONE) :: hd l_s :: acc) xs (tl l_s) in
            String.concat (x :: aux [] xs l)
      end
    in
      fun sprintf0 s_pat = s_pat
      fun sprintf1 s_pat s1 = sprintf s_pat [s1]
      fun sprintf2 s_pat s1 s2 = sprintf s_pat [s1, s2]
      fun sprintf3 s_pat s1 s2 s3 = sprintf s_pat [s1, s2, s3]
      fun sprintf4 s_pat s1 s2 s3 s4 = sprintf s_pat [s1, s2, s3, s4]
      fun sprintf5 s_pat s1 s2 s3 s4 s5 = sprintf s_pat [s1, s2, s3, s4, s5]
    end
  end

  structure String = struct
    val concat = String.concatWith
  end

  structure Sys = struct
    val isDirectory2 = SOME o File.is_dir o Path.explode handle ERROR _ => K NONE
  end

  structure To = struct
    fun nat f = Int.toString o f
  end

  fun outFile1 _ _ = tap (fn _ => warning "not implemented") NONE
  fun outStand1 f = outFile1 f ""
end
\<close>

ML_file \<open>../generated/c_ast.ML\<close>

text \<open> Note that the purpose of \<^dir>\<open>../generated\<close> is to host any generated
files of the Isabelle/C project. It contains among others:

  \<^item> \<^file>\<open>../generated/c_ast.ML\<close> representing the Abstract Syntax Tree of C,
  which has just been loaded.
  
  \<^item> \<^file>\<open>../generated/c_grammar_fun.grm\<close> is a generated file not used by the
  project, except for further generating \<^file>\<open>../generated/c_grammar_fun.grm.sig\<close>
  and \<^file>\<open>../generated/c_grammar_fun.grm.sml\<close>. Its physical presence in the
  directory is actually not necessary, but has been kept for informative documentation purposes. It
  represents the basis point of our SML grammar file, generated by an initial Haskell grammar file
  (namely
  \<^url>\<open>https://github.com/visq/language-c/blob/master/src/Language/C/Parser/Parser.y\<close>)
  using a modified version of Happy.

  \<^item> \<^file>\<open>../generated/c_grammar_fun.grm.sig\<close> and
  \<^file>\<open>../generated/c_grammar_fun.grm.sml\<close> are the two files generated from
  \<^file>\<open>../generated/c_grammar_fun.grm\<close> with a modified version of ML-Yacc. This
  last comes from MLton source in \<^dir>\<open>../../src_ext/mlton\<close>, see for example
  \<^dir>\<open>../../src_ext/mlton/mlyacc\<close>.
\<close>

text \<open> For the case of \<^file>\<open>../generated/c_ast.ML\<close>, it is actually not
mandatory to have a ``physical'' representation of the file in \<^dir>\<open>../generated\<close>:
it could be generated ``on-the-fly'' with \<^theory_text>\<open>code_reflect\<close> and immediately
loaded: Citadelle has an option to choose between the two
tasks~\cite{DBLP:journals/afp/TuongW15}.\<^footnote>\<open>\<^url>\<open>https://gitlri.lri.fr/ftuong/citadelle-devel\<close>\<close>\<close>

text \<open> After loading the AST, it is possible in Citadelle to do various meta-programming
renaming, such as the one depicted in the next command. Actually, its content is explicitly included
here by hand since we decided to manually load the AST using the above
\<^theory_text>\<open>ML_file\<close> command. (Otherwise, one can automatically execute the overall
generation and renaming tasks in Citadelle without resorting to a manual copying-pasting.)\<close>

ML \<comment> \<open>\<^file>\<open>../generated/c_ast.ML\<close>\<close> \<open>
structure C_Ast =
struct
  val Position = C_Ast.position val NoPosition = C_Ast.noPosition val BuiltinPosition = C_Ast.builtinPosition val InternalPosition = C_Ast.internalPosition val Name = C_Ast.name val OnlyPos = C_Ast.onlyPos val NodeInfo = C_Ast.nodeInfo val AnonymousRef = C_Ast.anonymousRef val NamedRef = C_Ast.namedRef val CChar = C_Ast.cChar val CChars = C_Ast.cChars val DecRepr = C_Ast.decRepr val HexRepr = C_Ast.hexRepr val OctalRepr = C_Ast.octalRepr val FlagUnsigned = C_Ast.flagUnsigned val FlagLong = C_Ast.flagLong val FlagLongLong = C_Ast.flagLongLong val FlagImag = C_Ast.flagImag val CFloat = C_Ast.cFloat val Flags = C_Ast.flags val CInteger = C_Ast.cInteger val CAssignOp = C_Ast.cAssignOp val CMulAssOp = C_Ast.cMulAssOp val CDivAssOp = C_Ast.cDivAssOp val CRmdAssOp = C_Ast.cRmdAssOp val CAddAssOp = C_Ast.cAddAssOp val CSubAssOp = C_Ast.cSubAssOp val CShlAssOp = C_Ast.cShlAssOp val CShrAssOp = C_Ast.cShrAssOp val CAndAssOp = C_Ast.cAndAssOp val CXorAssOp = C_Ast.cXorAssOp val COrAssOp = C_Ast.cOrAssOp val CMulOp = C_Ast.cMulOp val CDivOp = C_Ast.cDivOp val CRmdOp = C_Ast.cRmdOp val CAddOp = C_Ast.cAddOp val CSubOp = C_Ast.cSubOp val CShlOp = C_Ast.cShlOp val CShrOp = C_Ast.cShrOp val CLeOp = C_Ast.cLeOp val CGrOp = C_Ast.cGrOp val CLeqOp = C_Ast.cLeqOp val CGeqOp = C_Ast.cGeqOp val CEqOp = C_Ast.cEqOp val CNeqOp = C_Ast.cNeqOp val CAndOp = C_Ast.cAndOp val CXorOp = C_Ast.cXorOp val COrOp = C_Ast.cOrOp val CLndOp = C_Ast.cLndOp val CLorOp = C_Ast.cLorOp val CPreIncOp = C_Ast.cPreIncOp val CPreDecOp = C_Ast.cPreDecOp val CPostIncOp = C_Ast.cPostIncOp val CPostDecOp = C_Ast.cPostDecOp val CAdrOp = C_Ast.cAdrOp val CIndOp = C_Ast.cIndOp val CPlusOp = C_Ast.cPlusOp val CMinOp = C_Ast.cMinOp val CCompOp = C_Ast.cCompOp val CNegOp = C_Ast.cNegOp val CAuto = C_Ast.cAuto val CRegister = C_Ast.cRegister val CStatic = C_Ast.cStatic val CExtern = C_Ast.cExtern val CTypedef = C_Ast.cTypedef val CThread = C_Ast.cThread val CInlineQual = C_Ast.cInlineQual val CNoreturnQual = C_Ast.cNoreturnQual val CStructTag = C_Ast.cStructTag val CUnionTag = C_Ast.cUnionTag val CIntConst = C_Ast.cIntConst val CCharConst = C_Ast.cCharConst val CFloatConst = C_Ast.cFloatConst val CStrConst = C_Ast.cStrConst val CStrLit = C_Ast.cStrLit val CFunDef = C_Ast.cFunDef val CDecl = C_Ast.cDecl val CStaticAssert = C_Ast.cStaticAssert val CDeclr = C_Ast.cDeclr val CPtrDeclr = C_Ast.cPtrDeclr val CArrDeclr = C_Ast.cArrDeclr val CFunDeclr = C_Ast.cFunDeclr val CNoArrSize = C_Ast.cNoArrSize val CArrSize = C_Ast.cArrSize val CLabel = C_Ast.cLabel val CCase = C_Ast.cCase val CCases = C_Ast.cCases val CDefault = C_Ast.cDefault val CExpr = C_Ast.cExpr val CCompound = C_Ast.cCompound val CIf = C_Ast.cIf val CSwitch = C_Ast.cSwitch val CWhile = C_Ast.cWhile val CFor = C_Ast.cFor val CGoto = C_Ast.cGoto val CGotoPtr = C_Ast.cGotoPtr val CCont = C_Ast.cCont val CBreak = C_Ast.cBreak val CReturn = C_Ast.cReturn val CAsm = C_Ast.cAsm val CAsmStmt = C_Ast.cAsmStmt val CAsmOperand = C_Ast.cAsmOperand val CBlockStmt = C_Ast.cBlockStmt val CBlockDecl = C_Ast.cBlockDecl val CNestedFunDef = C_Ast.cNestedFunDef val CStorageSpec = C_Ast.cStorageSpec val CTypeSpec = C_Ast.cTypeSpec val CTypeQual = C_Ast.cTypeQual val CFunSpec = C_Ast.cFunSpec val CAlignSpec = C_Ast.cAlignSpec val CVoidType = C_Ast.cVoidType val CCharType = C_Ast.cCharType val CShortType = C_Ast.cShortType val CIntType = C_Ast.cIntType val CLongType = C_Ast.cLongType val CFloatType = C_Ast.cFloatType val CDoubleType = C_Ast.cDoubleType val CSignedType = C_Ast.cSignedType val CUnsigType = C_Ast.cUnsigType val CBoolType = C_Ast.cBoolType val CComplexType = C_Ast.cComplexType val CInt128Type = C_Ast.cInt128Type val CSUType = C_Ast.cSUType val CEnumType = C_Ast.cEnumType val CTypeDef = C_Ast.cTypeDef val CTypeOfExpr = C_Ast.cTypeOfExpr val CTypeOfType = C_Ast.cTypeOfType val CAtomicType = C_Ast.cAtomicType val CConstQual = C_Ast.cConstQual val CVolatQual = C_Ast.cVolatQual val CRestrQual = C_Ast.cRestrQual val CAtomicQual = C_Ast.cAtomicQual val CAttrQual = C_Ast.cAttrQual val CNullableQual = C_Ast.cNullableQual val CNonnullQual = C_Ast.cNonnullQual val CAlignAsType = C_Ast.cAlignAsType val CAlignAsExpr = C_Ast.cAlignAsExpr val CStruct = C_Ast.cStruct val CEnum = C_Ast.cEnum val CInitExpr = C_Ast.cInitExpr val CInitList = C_Ast.cInitList val CArrDesig = C_Ast.cArrDesig val CMemberDesig = C_Ast.cMemberDesig val CRangeDesig = C_Ast.cRangeDesig val CAttr = C_Ast.cAttr val CComma = C_Ast.cComma val CAssign = C_Ast.cAssign val CCond = C_Ast.cCond val CBinary = C_Ast.cBinary val CCast = C_Ast.cCast val CUnary = C_Ast.cUnary val CSizeofExpr = C_Ast.cSizeofExpr val CSizeofType = C_Ast.cSizeofType val CAlignofExpr = C_Ast.cAlignofExpr val CAlignofType = C_Ast.cAlignofType val CComplexReal = C_Ast.cComplexReal val CComplexImag = C_Ast.cComplexImag val CIndex = C_Ast.cIndex val CCall = C_Ast.cCall val CMember = C_Ast.cMember val CVar = C_Ast.cVar val CConst = C_Ast.cConst val CCompoundLit = C_Ast.cCompoundLit val CGenericSelection = C_Ast.cGenericSelection val CStatExpr = C_Ast.cStatExpr val CLabAddrExpr = C_Ast.cLabAddrExpr val CBuiltinExpr = C_Ast.cBuiltinExpr val CBuiltinVaArg = C_Ast.cBuiltinVaArg val CBuiltinOffsetOf = C_Ast.cBuiltinOffsetOf val CBuiltinTypesCompatible = C_Ast.cBuiltinTypesCompatible val CDeclExt = C_Ast.cDeclExt val CFDefExt = C_Ast.cFDefExt val CAsmExt = C_Ast.cAsmExt val CTranslUnit = C_Ast.cTranslUnit
  open C_Ast
end
\<close>

subsection \<open>Basic Aliases and Initialization of the Haskell Library\<close>

ML \<comment> \<open>\<^file>\<open>../generated/c_ast.ML\<close>\<close> \<open>
structure C_Ast =
struct
type class_Pos = Position.T * Position.T
(**)
type NodeInfo = C_Ast.nodeInfo
type CStorageSpec = NodeInfo C_Ast.cStorageSpecifier
type CFunSpec = NodeInfo C_Ast.cFunctionSpecifier
type CConst = NodeInfo C_Ast.cConstant
type 'a CInitializerList = ('a C_Ast.cPartDesignator List.list * 'a C_Ast.cInitializer) List.list
type CTranslUnit = NodeInfo C_Ast.cTranslationUnit
type CExtDecl = NodeInfo C_Ast.cExternalDeclaration
type CFunDef = NodeInfo C_Ast.cFunctionDef
type CDecl = NodeInfo C_Ast.cDeclaration
type CDeclr = NodeInfo C_Ast.cDeclarator
type CDerivedDeclr = NodeInfo C_Ast.cDerivedDeclarator
type CArrSize = NodeInfo C_Ast.cArraySize
type CStat = NodeInfo C_Ast.cStatement
type CAsmStmt = NodeInfo C_Ast.cAssemblyStatement
type CAsmOperand = NodeInfo C_Ast.cAssemblyOperand
type CBlockItem = NodeInfo C_Ast.cCompoundBlockItem
type CDeclSpec = NodeInfo C_Ast.cDeclarationSpecifier
type CTypeSpec = NodeInfo C_Ast.cTypeSpecifier
type CTypeQual = NodeInfo C_Ast.cTypeQualifier
type CAlignSpec = NodeInfo C_Ast.cAlignmentSpecifier
type CStructUnion = NodeInfo C_Ast.cStructureUnion
type CEnum = NodeInfo C_Ast.cEnumeration
type CInit = NodeInfo C_Ast.cInitializer
type CInitList = NodeInfo CInitializerList
type CDesignator = NodeInfo C_Ast.cPartDesignator
type CAttr = NodeInfo C_Ast.cAttribute
type CExpr = NodeInfo C_Ast.cExpression
type CBuiltin = NodeInfo C_Ast.cBuiltinThing
type CStrLit = NodeInfo C_Ast.cStringLiteral
(**)
type ClangCVersion = C_Ast.clangCVersion
type Ident = C_Ast.ident
type Position = C_Ast.positiona
type PosLength = Position * int
type Name = C_Ast.namea
type Bool = bool
type CString = C_Ast.cString
type CChar = C_Ast.cChar
type CInteger = C_Ast.cInteger
type CFloat = C_Ast.cFloat
type CStructTag = C_Ast.cStructTag
type CUnaryOp = C_Ast.cUnaryOp
type 'a CStringLiteral = 'a C_Ast.cStringLiteral
type 'a CConstant = 'a C_Ast.cConstant
type ('a, 'b) Either = ('a, 'b) C_Ast.either
type CIntRepr = C_Ast.cIntRepr
type CIntFlag = C_Ast.cIntFlag
type CAssignOp = C_Ast.cAssignOp
type Comment = C_Ast.comment
(**)
type 'a Reversed = 'a
datatype CDeclrR = CDeclrR0 of C_Ast.ident C_Ast.optiona
                             * NodeInfo C_Ast.cDerivedDeclarator list Reversed
                             * NodeInfo C_Ast.cStringLiteral C_Ast.optiona
                             * NodeInfo C_Ast.cAttribute list
                             * NodeInfo
type 'a Maybe = 'a C_Ast.optiona
datatype 'a Located = Located of 'a * (Position * (Position * int))
(**)
fun CDeclrR ide l s a n = CDeclrR0 (ide, l, s, a, n)
val reverse = rev
val Nothing = C_Ast.None
val Just = C_Ast.Some
val False = false
val True = true
val Ident = C_Ast.Ident0
(**)
val CDecl_flat = fn l1 => C_Ast.CDecl l1 o map (fn (a, b, c) => ((a, b), c))
fun flat3 (a, b, c) = ((a, b), c)
fun maybe def f = fn C_Ast.None => def | C_Ast.Some x => f x 
val Reversed = I
(**)
val From_string =
    C_Ast.SS_base
  o C_Ast.ST
  o implode
  o map (fn s => \<comment> \<open>prevent functions in \<^file>\<open>~~/src/HOL/String.thy\<close> of raising additional errors
                     (e.g., see the ML code associated to \<^term>\<open>String.asciis_of_literal\<close>)\<close>
          if Symbol.is_char s then s
          else if Symbol.is_utf8 s then translate_string (fn c => "\\" ^ string_of_int (ord c)) s
          else s)
  o Symbol.explode
val From_char_hd = hd o C_Ast.explode
(**)
val Namea = C_Ast.name
(**)
open C_Ast
fun flip f b a = f a b
open Basic_Library
end
\<close>

ML\<open>

signature C11_AST_LIB =
  sig
    val fold_either: ('a -> 'b -> 'c) -> ('d -> 'b -> 'c) -> ('a, 'd) C_Ast.either -> 'b -> 'c
    val fold_ident: 'a -> (string -> 'a -> 'b -> 'c) -> C_Ast.ident -> 'b -> 'c
    val fold_optiona: ('a -> 'b -> 'b) -> 'a C_Ast.optiona -> 'b -> 'b

    val toString_cBinaryOp: C_Ast.cBinaryOp -> string
    val toString_cIntFlag: C_Ast.cIntFlag -> string
    val toString_cIntRepr: C_Ast.cIntRepr -> string
    val toString_cUnaryOp: C_Ast.cUnaryOp -> string
    val toString_cAssignOp: C_Ast.cAssignOp -> string

    val fold_cAttribute: (string -> 'a -> 'b -> 'b) -> 'a C_Ast.cAttribute -> 'b -> 'b
    val fold_cBuiltinThing: (string -> 'a -> 'b -> 'b) -> 'a C_Ast.cBuiltinThing -> 'b -> 'b
    val fold_cConstant: (string -> 'a -> 'b -> 'c) -> 'a C_Ast.cConstant -> 'b -> 'c
    val fold_cDeclaration: (string -> 'a -> 'b -> 'b) -> 'a C_Ast.cDeclaration -> 'b -> 'b
    val fold_cEnumeration: (string -> 'a -> 'b -> 'b) -> 'a C_Ast.cEnumeration -> 'b -> 'b
    val fold_cExpression: (string -> 'a -> 'b -> 'b) -> 'a C_Ast.cExpression -> 'b -> 'b
    val fold_cInitializer: (string -> 'a -> 'b -> 'b) -> 'a C_Ast.cInitializer -> 'b -> 'b
    val fold_cInteger: (string -> 'a -> 'b) -> C_Ast.cInteger -> 'a -> 'b
    val fold_cPartDesignator: (string -> 'a -> 'b -> 'b) -> 'a C_Ast.cPartDesignator -> 'b -> 'b
    val fold_cStatement: (string -> 'a -> 'b -> 'b) -> 'a C_Ast.cStatement -> 'b -> 'b
    val fold_cStringLiteral: (string -> 'a -> 'b -> 'c) -> 'a C_Ast.cStringLiteral -> 'b -> 'c
    val fold_cTypeQualifier: (string -> 'a -> 'b -> 'b) -> 'a C_Ast.cTypeQualifier -> 'b -> 'b
    val fold_cTypeSpecifier: (string -> 'a -> 'b -> 'b) -> 'a C_Ast.cTypeSpecifier -> 'b -> 'b
  end


structure C11_Ast_Lib : C11_AST_LIB =
struct 
local open  C_Ast in

fun fold_optiona _ None st = st | fold_optiona g (Some a) st = g a st;
   fun fold_either g1 _ (Left a) st = g1 a st
      |fold_either _ g2 (Right a) st = g2 a st


fun toString_cIntFlag FlagUnsigned0 = "FlagUnsigned0"
  | toString_cIntFlag FlagLong0     = "FlagLong0"
  | toString_cIntFlag FlagLongLong0 = "FlagLongLong0"
  | toString_cIntFlag FlagImag0     = "FlagImag0"

fun toString_cIntRepr DecRepr0      = "DecRepr0"
  | toString_cIntRepr HexRepr0      = "HexRepr0"
  | toString_cIntRepr OctalRepr0    = "OctalRepr0"


fun toString_cUnaryOp CPreIncOp0  = "CPreIncOp0"
  | toString_cUnaryOp CPreDecOp0  = "CPreDecOp0"
  | toString_cUnaryOp CPostIncOp0 = "CPostIncOp0"
  | toString_cUnaryOp CPostDecOp0 = "CPostDecOp0"
  | toString_cUnaryOp CAdrOp0     = "CAdrOp0"
  | toString_cUnaryOp CIndOp0     = "CIndOp0"
  | toString_cUnaryOp CPlusOp0    = "CPlusOp0"
  | toString_cUnaryOp CMinOp0     = "CMinOp0"
  | toString_cUnaryOp CCompOp0    = "CCompOp0"
  | toString_cUnaryOp CNegOp0     = "CNegOp0"

fun toString_cAssignOp CAssignOp0 = "CAssignOp0"
  | toString_cAssignOp CMulAssOp0 = "CMulAssOp0"
  | toString_cAssignOp CDivAssOp0 = "CDivAssOp0"
  | toString_cAssignOp CRmdAssOp0 = "CRmdAssOp0"
  | toString_cAssignOp CAddAssOp0 = "CAddAssOp0"
  | toString_cAssignOp CSubAssOp0 = "CSubAssOp0"
  | toString_cAssignOp CShlAssOp0 = "CShlAssOp0"
  | toString_cAssignOp CShrAssOp0 = "CShrAssOp0"
  | toString_cAssignOp CAndAssOp0 = "CAndAssOp0"
  | toString_cAssignOp CXorAssOp0 = "CXorAssOp0"
  | toString_cAssignOp COrAssOp0  = "COrAssOp0"

fun toString_cBinaryOp CMulOp0   = "CMulOp0"
  | toString_cBinaryOp CDivOp0   = "CDivOp0"
  | toString_cBinaryOp CRmdOp0   = "CRmdOp0"
  | toString_cBinaryOp CAddOp0   = "CAddOp0"
  | toString_cBinaryOp CSubOp0   = "CSubOp0"
  | toString_cBinaryOp CShlOp0   = "CShlOp0"
  | toString_cBinaryOp CShrOp0   = "CShrOp0"
  | toString_cBinaryOp CLeOp0    = "CLeOp0"
  | toString_cBinaryOp CGrOp0    = "CGrOp0"
  | toString_cBinaryOp CLeqOp0   = "CLeqOp0"
  | toString_cBinaryOp CGeqOp0   = "CGeqOp0"
  | toString_cBinaryOp CEqOp0    = "CEqOp0"
  | toString_cBinaryOp CNeqOp0   = "CNeqOp0"
  | toString_cBinaryOp CAndOp0   = "CAndOp0"
  | toString_cBinaryOp CXorOp0   = "CXorOp0"
  | toString_cBinaryOp COrOp0    = "COrOp0"
  | toString_cBinaryOp CLndOp0   = "CLndOp0"
  | toString_cBinaryOp CLorOp0   = "CLorOp0"


fun fold_cInteger  h (CInteger0 (_: int, _: cIntRepr, _:  cIntFlag flags)) st = st |> h "CInteger0" 
    (* could be done much more *)

fun  fold_cConstant g (CIntConst0 (_: cInteger, a))  st = g "CIntConst0" a st 
   | fold_cConstant g (CCharConst0 (_ : cChar, a))   st = g "CCharConst0" a st  
   | fold_cConstant g (CFloatConst0 (_ : cFloat, a)) st = g "CFloatConst0" a st  
   | fold_cConstant g (CStrConst0 (_ : cString, a))  st = g "CStrConst0" a st 

fun fold_ident a g (Ident0(_: abr_string, _ : int, _: nodeInfo (* ignored here !!! *))) st = 
                         st |> g "Ident0" a
    (* there could be done much much more !!! *)
    (* that ident contains nodeinfo instead of 'a is an error in my view. *)

fun fold_cStringLiteral g (CStrLit0(_:cString, a)) st =  st |> g "CStrLit0" a 


fun  fold_cTypeSpecifier g (CAtomicType0 (decl : 'a cDeclaration, a)) st = 
                                  st |> fold_cDeclaration g decl |> g "CAtomicType0" a
   | fold_cTypeSpecifier g (CBoolType0 a)     st = g "CBoolType0" a st
   | fold_cTypeSpecifier g (CCharType0 a)     st = g "CCharType0" a st
   | fold_cTypeSpecifier g (CComplexType0 a)  st = g "CComplexType0" a st
   | fold_cTypeSpecifier g (CDoubleType0 a)   st = g "CDoubleType0" a st
   | fold_cTypeSpecifier g (CEnumType0(_: 'a cEnumeration, a)) st = g "CEnumType0" a st
   | fold_cTypeSpecifier g (CFloatType0 a)    st = g "CFloatType0" a st
   | fold_cTypeSpecifier g (CInt128Type0 a)   st = g "CInt128Type0" a st
   | fold_cTypeSpecifier g (CIntType0 a)      st = g "CIntType0" a st
   | fold_cTypeSpecifier g (CLongType0 a)     st = g "CLongType0" a st
   | fold_cTypeSpecifier g (CSUType0 (_: 'a cStructureUnion, a))  st = g "CSUType0" a st
   | fold_cTypeSpecifier g (CShortType0 a)    st = g "CShortType0" a st
   | fold_cTypeSpecifier g (CSignedType0 a)   st = g "CSignedType0" a st
   | fold_cTypeSpecifier g (CTypeDef0 (id:ident, a)) st = 
                                              st |>  fold_ident a g id
                                                 |>  g "CTypeDef0" a 
   | fold_cTypeSpecifier g (CTypeOfExpr0 (ex: 'a cExpression, a)) st = 
                                  st |> fold_cExpression g ex |> g "CTypeOfExpr0" a
   | fold_cTypeSpecifier g (CTypeOfType0 (decl: 'a cDeclaration, a)) st = 
                                  st |> fold_cDeclaration g decl |> g "CTypeOfType0" a
   | fold_cTypeSpecifier g (CUnsigType0 a)    st = st |> g "CUnsigType0" a
   | fold_cTypeSpecifier g (CVoidType0 a)     st = st |> g "CVoidType0"  a


and  fold_cTypeQualifier g (CAtomicQual0 a) st = g "CAtomicQual0" a st
   | fold_cTypeQualifier g (CAttrQual0 (CAttr0 (id: ident,
                                                eL: 'a cExpression list,
                                                a))) st = st |> fold_ident a g id
                                                             |> fold(fold_cExpression g) eL
                                                             |> g "CAttrQual0" a
   | fold_cTypeQualifier g (CConstQual0 a) st    = st |> g "CConstQual0" a
   | fold_cTypeQualifier g (CNonnullQual0 a) st  = st |> g "CNonnullQual0" a 
   | fold_cTypeQualifier g (CNullableQual0 a) st = st |> g "CNullableQual0" a
   | fold_cTypeQualifier g (CRestrQual0 a) st    = st |> g "CRestrQual0" a
   | fold_cTypeQualifier g (CVolatQual0 a) st    = st |> g "CVolatQual0" a 

and  fold_cStatement g (CLabel0(id:ident, s:'a cStatement, 
                                aL: 'a cAttribute list, a)) st= 
                                  st |> fold_ident a g id
                                     |> fold_cStatement g s
                                     |> fold(fold_cAttribute g) aL
                                     |> g "CLabel0" a 
   | fold_cStatement g (CCase0(ex: 'a cExpression, 
                               stmt: 'a cStatement, a)) st    = 
                                  st |> fold_cExpression g ex
                                     |> fold_cStatement g stmt
                                     |> g "CCase0" a
   | fold_cStatement g (CCases0(ex1: 'a cExpression, 
                                ex2: 'a cExpression,
                                stmt:'a cStatement, a)) st   = 
                                  st |> fold_cExpression g ex1
                                     |> fold_cExpression g ex2
                                     |> fold_cStatement g stmt
                                     |> g "CCases0" a
   | fold_cStatement g (CDefault0(stmt:'a cStatement, a)) st  = 
                                  st |> fold_cStatement g stmt
                                     |> g "CDefault0" a
   | fold_cStatement g (CExpr0(ex_opt:'a cExpression optiona, a)) st = 
                                  st |> fold_optiona (fold_cExpression g) ex_opt
                                     |> g "CExpr0" a
   | fold_cStatement g (CCompound0(idS : ident list, 
                                   _: 'a cCompoundBlockItem list, a)) st = (* TODO *)
                                  st |> fold(fold_ident a g) idS
                                     |> g "CCompound0" a
   | fold_cStatement g (CIf0(ex1:'a cExpression,stmt: 'a cStatement, 
                        stmt_opt: 'a cStatement optiona, a)) st = 
                                  st |> fold_cExpression g ex1
                                     |> fold_cStatement g stmt
                                     |> fold_optiona (fold_cStatement g) stmt_opt
                                     |> g "CIf0" a
   | fold_cStatement g (CSwitch0(ex1:'a cExpression, 
                                 stmt: 'a cStatement, a)) st = 
                                  st |> fold_cExpression g ex1
                                     |> fold_cStatement g stmt
                                     |> g "CSwitch0" a
   | fold_cStatement g (CWhile0(ex1:'a cExpression, 
                                stmt: 'a cStatement, b: bool, a)) st = 
                                  st |> fold_cExpression g ex1
                                     |> fold_cStatement g stmt
                                     |> g ("CWhile0 "^((b ? (K "F")) "T")) a
   | fold_cStatement g (CFor0(ex0:('a cExpression optiona, 'a cDeclaration) either, 
                              ex1_opt: 'a cExpression optiona, 
                              ex2_opt: 'a cExpression optiona,
                              stmt: 'a cStatement, a)) st = 
                                  st |> fold_either (fold_optiona (fold_cExpression g)) 
                                                    (fold_cDeclaration g) ex0
                                     |> fold_optiona (fold_cExpression g) ex1_opt
                                     |> fold_optiona (fold_cExpression g) ex2_opt
                                     |> fold_cStatement g stmt
                                     |> g "CFor0" a
   | fold_cStatement g (CGoto0(id: ident, a)) st = st |>  fold_ident a g id
                                                      |> g "CGoto0" a
   | fold_cStatement g (CGotoPtr0(ex1:'a cExpression, a)) st = 
                                  st |> fold_cExpression g ex1 |> g "CGotoPtr0" a
   | fold_cStatement g (CCont0 a) st  = st |> g "CCont0" a
   | fold_cStatement g (CBreak0 a) st = st |> g "CBreak0" a
   | fold_cStatement g (CReturn0 (ex:'a cExpression optiona,a)) st = 
                                  st |> fold_optiona (fold_cExpression g) ex |> g "CReturn0" a
   | fold_cStatement g (CAsm0(_: 'a cAssemblyStatement, a)) st = 
                                  st |> g "CAsm0" a

and fold_cExpression g (CComma0 (eL:'a cExpression list, a)) st = 
                                 st |> fold(fold_cExpression g) eL |> g "CComma0" a
  | fold_cExpression g (CAssign0(aop:cAssignOp, 
                                 ex1:'a cExpression,
                                 ex2:'a cExpression,a)) st = 
                                  st |> fold_cExpression g ex1
                                     |> fold_cExpression g ex2 
                                     |> g ("CAssign0 "^toString_cAssignOp aop) a
  | fold_cExpression g (CCond0(  ex1:'a cExpression, 
                                 ex2opt: 'a cExpression optiona, (* bescheuert ! Wieso option ?*) 
                                 ex3: 'a cExpression,a)) st = 
                                  st |> fold_cExpression g ex1 
                                     |> fold_optiona (fold_cExpression g) ex2opt
                                     |> fold_cExpression g ex3 |> g "CCond0" a
  | fold_cExpression g (CBinary0(bop: cBinaryOp, ex1: 'a cExpression,ex2: 'a cExpression, a)) st =
                                 st |> fold_cExpression g ex1 
                                    |> fold_cExpression g ex2 
                                    |> g ("CBinary0 "^toString_cBinaryOp bop) a 
  | fold_cExpression g (CCast0(decl:'a cDeclaration, ex: 'a cExpression, a)) st = 
                                 st |> fold_cExpression g ex 
                                    |> fold_cDeclaration g decl
                                    |> g "CCast0" a
  | fold_cExpression g (CUnary0(unop:cUnaryOp, ex: 'a cExpression, a)) st = 
                                 st |> fold_cExpression g ex 
                                    |> g ("CUnary0 "^toString_cUnaryOp unop) a
  | fold_cExpression g (CSizeofExpr0(ex:'a cExpression, a)) st = 
                                 st |> fold_cExpression g ex    |> g "CSizeofExpr0" a
  | fold_cExpression g (CSizeofType0(decl:'a cDeclaration,a)) st = 
                                 st |> fold_cDeclaration g decl |> g "CSizeofType0" a
  | fold_cExpression g (CAlignofExpr0(ex:'a cExpression, a)) st = 
                                 st |> fold_cExpression g ex    |> g "CAlignofExpr0" a
  | fold_cExpression g (CAlignofType0(decl:'a cDeclaration, a)) st = 
                                 st |> fold_cDeclaration g decl |> g "CAlignofType0" a
  | fold_cExpression g (CComplexReal0(ex:'a cExpression, a)) st = 
                                 st |> fold_cExpression g ex |> g "CComplexReal0" a
  | fold_cExpression g (CComplexImag0(ex:'a cExpression, a)) st = 
                                 st |> fold_cExpression g ex |> g "CComplexImag0" a
  | fold_cExpression g (CIndex0(ex1:'a cExpression, ex2: 'a cExpression, a)) st =
                                 st |> fold_cExpression g ex1 
                                    |> fold_cExpression g ex2 
                                    |> g "CIndex0" a
  | fold_cExpression g (CCall0(ex:'a cExpression, argS: 'a cExpression list, a)) st =
                                 st |> fold_cExpression g ex 
                                    |> fold (fold_cExpression g) argS 
                                    |> g "CCall0" a
  | fold_cExpression g (CMember0(ex:'a cExpression, id:ident, b, a)) st =
                                 st |> fold_cExpression g ex 
                                    |> fold_ident a g id
                                    |> g ("CMember0"^((b ? (K "F")) "T")) a 
  | fold_cExpression g (CVar0(id:ident,a)) st = st |> fold_ident a g id |> g "CVar0" a
  | fold_cExpression g (CConst0(cc:'a cConstant)) st  = st |> fold_cConstant g cc
  | fold_cExpression g (CCompoundLit0(decl:'a cDeclaration,
                                      eqn: ('a cPartDesignator list * 'a cInitializer) list, a)) st =
                                 st  |> fold(fn(S,init) => 
                                             fn st => st |> fold(fold_cPartDesignator g) S
                                                         |> fold_cInitializer g init) eqn 
                                     |> fold_cDeclaration g decl 
                                     |> g "CCompoundLit0" a 
  | fold_cExpression g (CGenericSelection0(ex:'a cExpression, 
                                           eqn: ('a cDeclaration optiona*'a cExpression)list,a)) st =
                                 st |> fold_cExpression g ex 
                                    |> fold (fn (d,ex) => 
                                             fn st => st |> fold_optiona (fold_cDeclaration g) d  
                                                         |> fold_cExpression g ex) eqn  
                                    |> g "CGenericSelection0" a  
  | fold_cExpression g (CStatExpr0(stmt: 'a cStatement,a)) st =  
                                 st |> fold_cStatement g stmt |> g "CStatExpr0" a
  | fold_cExpression g (CLabAddrExpr0(id:ident,a)) st = 
                                 st |> fold_ident a g id |> g "CLabAddrExpr0" a 
  | fold_cExpression g (CBuiltinExpr0(X: 'a cBuiltinThing)) st = st |> fold_cBuiltinThing g X
  
and fold_cDeclaration g (CDecl0(_ : 'a cDeclarationSpecifier list, (* TODO *)
                                mkS: (('a cDeclarator optiona*'a cInitializer optiona) * 'a cExpression optiona) list,
                                a)) st = 
                                 st |> fold(fn ((d_o, init_o),ex_opt) =>
                                            fn st => st |> fold_optiona(fold_cInitializer g) init_o
                                                        |> fold_optiona(fold_cExpression g) ex_opt) mkS
                                    |> g "CDecl0" a
  | fold_cDeclaration g (CStaticAssert0(ex:'a cExpression, slit: 'a cStringLiteral, a)) st = 
                                 st |> fold_cExpression g ex 
                                    |> fold_cStringLiteral g slit
                                    |> g "CStaticAssert0" a

and fold_cBuiltinThing g (CBuiltinVaArg0(ex:'a cExpression,decl: 'a cDeclaration,a)) st = 
                                 st |> fold_cExpression g ex 
                                    |> fold_cDeclaration g decl
                                    |> g "CBuiltinVaArg0" a 
  | fold_cBuiltinThing g (CBuiltinOffsetOf0(d: 'a cDeclaration, _: 'a cPartDesignator list, a)) st = 
                                 st |> fold_cDeclaration g d 
                                    |> g "CBuiltinOffsetOf0" a 
  | fold_cBuiltinThing g (CBuiltinTypesCompatible0 (d1: 'a cDeclaration, d2: 'a cDeclaration,a)) st= 
                                 st |> fold_cDeclaration g d1
                                    |> fold_cDeclaration g d2 
                                    |> g "CBuiltinTypesCompatible0" a 

and  fold_cInitializer g (CInitExpr0(ex: 'a cExpression, a)) st = 
                                 st |> fold_cExpression g ex |> g "CInitExpr0" a 
   | fold_cInitializer g (CInitList0 (mms: ('a cPartDesignator list * 'a cInitializer) list,a)) st = 
                                 st |> fold(fn (a,b) => 
                                            fn st => st|> fold(fold_cPartDesignator g) a 
                                                       |> fold_cInitializer g b) mms
                                    |> g "CInitList0" a

and  fold_cPartDesignator g (CArrDesig0(ex: 'a cExpression, a)) st = 
                                 st |> fold_cExpression g ex |> g "CArrDesig0" a 
   | fold_cPartDesignator g (CMemberDesig0(id: ident, a)) st = 
                                 st |> fold_ident a g id |> g "CMemberDesig0" a 
   | fold_cPartDesignator g (CRangeDesig0(ex1: 'a cExpression, ex2: 'a cExpression, a)) st = 
                                 st |> fold_cExpression g ex1
                                    |> fold_cExpression g ex2
                                    |> g "CRangeDesig0" a 

and fold_cAttribute g (CAttr0(id: ident, exS: 'a cExpression list, a)) st =
                                 st |> fold_ident a g id
                                    |> fold(fold_cExpression g) exS 
                                    |> g "CAttr0" a 

and fold_cEnumeration g (CEnum0 (ident_opt: ident optiona,
                              exS_opt: ((ident * 'a cExpression optiona) list) optiona,
                              attrS: 'a cAttribute list, a)) st = 
                                  st |> fold_optiona(fold_ident a g) ident_opt
                                     |> fold_optiona(fold(
                                            fn (id,ex_o) =>
                                            fn st => st |> fold_ident a g id
                                                        |> fold_optiona (fold_cExpression g) ex_o)) 
                                                     exS_opt   
                                     |> fold(fold_cAttribute g) attrS
                                     |> g "CEnum0" a 


end

end (*struct *)
\<close>
ML\<open>

signature C11_AST_LIB =
  sig
    val fold_either: ('a -> 'b -> 'c) -> ('d -> 'b -> 'c) -> ('a, 'd) C_Ast.either -> 'b -> 'c
    val fold_ident: 'a -> (string -> 'a -> 'b -> 'c) -> C_Ast.ident -> 'b -> 'c
    val fold_optiona: ('a -> 'b -> 'b) -> 'a C_Ast.optiona -> 'b -> 'b

    val toString_cBinaryOp: C_Ast.cBinaryOp -> string
    val toString_cIntFlag: C_Ast.cIntFlag -> string
    val toString_cIntRepr: C_Ast.cIntRepr -> string
    val toString_cUnaryOp: C_Ast.cUnaryOp -> string
    val toString_cAssignOp: C_Ast.cAssignOp -> string

    val fold_cAttribute: (string -> 'a -> 'b -> 'b) -> 'a C_Ast.cAttribute -> 'b -> 'b
    val fold_cBuiltinThing: (string -> 'a -> 'b -> 'b) -> 'a C_Ast.cBuiltinThing -> 'b -> 'b
    val fold_cConstant: (string -> 'a -> 'b -> 'c) -> 'a C_Ast.cConstant -> 'b -> 'c
    val fold_cDeclaration: (string -> 'a -> 'b -> 'b) -> 'a C_Ast.cDeclaration -> 'b -> 'b
    val fold_cEnumeration: (string -> 'a -> 'b -> 'b) -> 'a C_Ast.cEnumeration -> 'b -> 'b
    val fold_cExpression: (string -> 'a -> 'b -> 'b) -> 'a C_Ast.cExpression -> 'b -> 'b
    val fold_cInitializer: (string -> 'a -> 'b -> 'b) -> 'a C_Ast.cInitializer -> 'b -> 'b
    val fold_cInteger: (string -> 'a -> 'b) -> C_Ast.cInteger -> 'a -> 'b
    val fold_cPartDesignator: (string -> 'a -> 'b -> 'b) -> 'a C_Ast.cPartDesignator -> 'b -> 'b
    val fold_cStatement: (string -> 'a -> 'b -> 'b) -> 'a C_Ast.cStatement -> 'b -> 'b
    val fold_cStringLiteral: (string -> 'a -> 'b -> 'c) -> 'a C_Ast.cStringLiteral -> 'b -> 'c
    val fold_cTypeQualifier: (string -> 'a -> 'b -> 'b) -> 'a C_Ast.cTypeQualifier -> 'b -> 'b
    val fold_cTypeSpecifier: (string -> 'a -> 'b -> 'b) -> 'a C_Ast.cTypeSpecifier -> 'b -> 'b
  end


structure C11_Ast_Lib : C11_AST_LIB =
struct 
local open  C_Ast in

fun fold_optiona _ None st = st | fold_optiona g (Some a) st = g a st;
   fun fold_either g1 _ (Left a) st = g1 a st
      |fold_either _ g2 (Right a) st = g2 a st


fun toString_cIntFlag FlagUnsigned0 = "FlagUnsigned0"
  | toString_cIntFlag FlagLong0     = "FlagLong0"
  | toString_cIntFlag FlagLongLong0 = "FlagLongLong0"
  | toString_cIntFlag FlagImag0     = "FlagImag0"

fun toString_cIntRepr DecRepr0      = "DecRepr0"
  | toString_cIntRepr HexRepr0      = "HexRepr0"
  | toString_cIntRepr OctalRepr0    = "OctalRepr0"


fun toString_cUnaryOp CPreIncOp0  = "CPreIncOp0"
  | toString_cUnaryOp CPreDecOp0  = "CPreDecOp0"
  | toString_cUnaryOp CPostIncOp0 = "CPostIncOp0"
  | toString_cUnaryOp CPostDecOp0 = "CPostDecOp0"
  | toString_cUnaryOp CAdrOp0     = "CAdrOp0"
  | toString_cUnaryOp CIndOp0     = "CIndOp0"
  | toString_cUnaryOp CPlusOp0    = "CPlusOp0"
  | toString_cUnaryOp CMinOp0     = "CMinOp0"
  | toString_cUnaryOp CCompOp0    = "CCompOp0"
  | toString_cUnaryOp CNegOp0     = "CNegOp0"

fun toString_cAssignOp CAssignOp0 = "CAssignOp0"
  | toString_cAssignOp CMulAssOp0 = "CMulAssOp0"
  | toString_cAssignOp CDivAssOp0 = "CDivAssOp0"
  | toString_cAssignOp CRmdAssOp0 = "CRmdAssOp0"
  | toString_cAssignOp CAddAssOp0 = "CAddAssOp0"
  | toString_cAssignOp CSubAssOp0 = "CSubAssOp0"
  | toString_cAssignOp CShlAssOp0 = "CShlAssOp0"
  | toString_cAssignOp CShrAssOp0 = "CShrAssOp0"
  | toString_cAssignOp CAndAssOp0 = "CAndAssOp0"
  | toString_cAssignOp CXorAssOp0 = "CXorAssOp0"
  | toString_cAssignOp COrAssOp0  = "COrAssOp0"

fun toString_cBinaryOp CMulOp0   = "CMulOp0"
  | toString_cBinaryOp CDivOp0   = "CDivOp0"
  | toString_cBinaryOp CRmdOp0   = "CRmdOp0"
  | toString_cBinaryOp CAddOp0   = "CAddOp0"
  | toString_cBinaryOp CSubOp0   = "CSubOp0"
  | toString_cBinaryOp CShlOp0   = "CShlOp0"
  | toString_cBinaryOp CShrOp0   = "CShrOp0"
  | toString_cBinaryOp CLeOp0    = "CLeOp0"
  | toString_cBinaryOp CGrOp0    = "CGrOp0"
  | toString_cBinaryOp CLeqOp0   = "CLeqOp0"
  | toString_cBinaryOp CGeqOp0   = "CGeqOp0"
  | toString_cBinaryOp CEqOp0    = "CEqOp0"
  | toString_cBinaryOp CNeqOp0   = "CNeqOp0"
  | toString_cBinaryOp CAndOp0   = "CAndOp0"
  | toString_cBinaryOp CXorOp0   = "CXorOp0"
  | toString_cBinaryOp COrOp0    = "COrOp0"
  | toString_cBinaryOp CLndOp0   = "CLndOp0"
  | toString_cBinaryOp CLorOp0   = "CLorOp0"


fun fold_cInteger  h (CInteger0 (_: int, _: cIntRepr, _:  cIntFlag flags)) st = st |> h "CInteger0" 
    (* could be done much more *)

fun  fold_cConstant g (CIntConst0 (_: cInteger, a))  st = g "CIntConst0" a st 
   | fold_cConstant g (CCharConst0 (_ : cChar, a))   st = g "CCharConst0" a st  
   | fold_cConstant g (CFloatConst0 (_ : cFloat, a)) st = g "CFloatConst0" a st  
   | fold_cConstant g (CStrConst0 (_ : cString, a))  st = g "CStrConst0" a st 

fun fold_ident a g (Ident0(_: abr_string, _ : int, _: nodeInfo (* ignored here !!! *))) st = 
                         st |> g "Ident0" a
    (* there could be done much much more !!! *)
    (* that ident contains nodeinfo instead of 'a is an error in my view. *)

fun fold_cStringLiteral g (CStrLit0(_:cString, a)) st =  st |> g "CStrLit0" a 


fun  fold_cTypeSpecifier g (CAtomicType0 (decl : 'a cDeclaration, a)) st = 
                                  st |> fold_cDeclaration g decl |> g "CAtomicType0" a
   | fold_cTypeSpecifier g (CBoolType0 a)     st = g "CBoolType0" a st
   | fold_cTypeSpecifier g (CCharType0 a)     st = g "CCharType0" a st
   | fold_cTypeSpecifier g (CComplexType0 a)  st = g "CComplexType0" a st
   | fold_cTypeSpecifier g (CDoubleType0 a)   st = g "CDoubleType0" a st
   | fold_cTypeSpecifier g (CEnumType0(_: 'a cEnumeration, a)) st = g "CEnumType0" a st
   | fold_cTypeSpecifier g (CFloatType0 a)    st = g "CFloatType0" a st
   | fold_cTypeSpecifier g (CInt128Type0 a)   st = g "CInt128Type0" a st
   | fold_cTypeSpecifier g (CIntType0 a)      st = g "CIntType0" a st
   | fold_cTypeSpecifier g (CLongType0 a)     st = g "CLongType0" a st
   | fold_cTypeSpecifier g (CSUType0 (_: 'a cStructureUnion, a))  st = g "CSUType0" a st
   | fold_cTypeSpecifier g (CShortType0 a)    st = g "CShortType0" a st
   | fold_cTypeSpecifier g (CSignedType0 a)   st = g "CSignedType0" a st
   | fold_cTypeSpecifier g (CTypeDef0 (id:ident, a)) st = 
                                              st |>  fold_ident a g id
                                                 |>  g "CTypeDef0" a 
   | fold_cTypeSpecifier g (CTypeOfExpr0 (ex: 'a cExpression, a)) st = 
                                  st |> fold_cExpression g ex |> g "CTypeOfExpr0" a
   | fold_cTypeSpecifier g (CTypeOfType0 (decl: 'a cDeclaration, a)) st = 
                                  st |> fold_cDeclaration g decl |> g "CTypeOfType0" a
   | fold_cTypeSpecifier g (CUnsigType0 a)    st = st |> g "CUnsigType0" a
   | fold_cTypeSpecifier g (CVoidType0 a)     st = st |> g "CVoidType0"  a


and  fold_cTypeQualifier g (CAtomicQual0 a) st = g "CAtomicQual0" a st
   | fold_cTypeQualifier g (CAttrQual0 (CAttr0 (id: ident,
                                                eL: 'a cExpression list,
                                                a))) st = st |> fold_ident a g id
                                                             |> fold(fold_cExpression g) eL
                                                             |> g "CAttrQual0" a
   | fold_cTypeQualifier g (CConstQual0 a) st    = st |> g "CConstQual0" a
   | fold_cTypeQualifier g (CNonnullQual0 a) st  = st |> g "CNonnullQual0" a 
   | fold_cTypeQualifier g (CNullableQual0 a) st = st |> g "CNullableQual0" a
   | fold_cTypeQualifier g (CRestrQual0 a) st    = st |> g "CRestrQual0" a
   | fold_cTypeQualifier g (CVolatQual0 a) st    = st |> g "CVolatQual0" a 

and  fold_cStatement g (CLabel0(id:ident, s:'a cStatement, 
                                aL: 'a cAttribute list, a)) st= 
                                  st |> fold_ident a g id
                                     |> fold_cStatement g s
                                     |> fold(fold_cAttribute g) aL
                                     |> g "CLabel0" a 
   | fold_cStatement g (CCase0(ex: 'a cExpression, 
                               stmt: 'a cStatement, a)) st    = 
                                  st |> fold_cExpression g ex
                                     |> fold_cStatement g stmt
                                     |> g "CCase0" a
   | fold_cStatement g (CCases0(ex1: 'a cExpression, 
                                ex2: 'a cExpression,
                                stmt:'a cStatement, a)) st   = 
                                  st |> fold_cExpression g ex1
                                     |> fold_cExpression g ex2
                                     |> fold_cStatement g stmt
                                     |> g "CCases0" a
   | fold_cStatement g (CDefault0(stmt:'a cStatement, a)) st  = 
                                  st |> fold_cStatement g stmt
                                     |> g "CDefault0" a
   | fold_cStatement g (CExpr0(ex_opt:'a cExpression optiona, a)) st = 
                                  st |> fold_optiona (fold_cExpression g) ex_opt
                                     |> g "CExpr0" a
   | fold_cStatement g (CCompound0(idS : ident list, 
                                   _: 'a cCompoundBlockItem list, a)) st = (* TODO *)
                                  st |> fold(fold_ident a g) idS
                                     |> g "CCompound0" a
   | fold_cStatement g (CIf0(ex1:'a cExpression,stmt: 'a cStatement, 
                        stmt_opt: 'a cStatement optiona, a)) st = 
                                  st |> fold_cExpression g ex1
                                     |> fold_cStatement g stmt
                                     |> fold_optiona (fold_cStatement g) stmt_opt
                                     |> g "CIf0" a
   | fold_cStatement g (CSwitch0(ex1:'a cExpression, 
                                 stmt: 'a cStatement, a)) st = 
                                  st |> fold_cExpression g ex1
                                     |> fold_cStatement g stmt
                                     |> g "CSwitch0" a
   | fold_cStatement g (CWhile0(ex1:'a cExpression, 
                                stmt: 'a cStatement, b: bool, a)) st = 
                                  st |> fold_cExpression g ex1
                                     |> fold_cStatement g stmt
                                     |> g ("CWhile0 "^((b ? (K "F")) "T")) a
   | fold_cStatement g (CFor0(ex0:('a cExpression optiona, 'a cDeclaration) either, 
                              ex1_opt: 'a cExpression optiona, 
                              ex2_opt: 'a cExpression optiona,
                              stmt: 'a cStatement, a)) st = 
                                  st |> fold_either (fold_optiona (fold_cExpression g)) 
                                                    (fold_cDeclaration g) ex0
                                     |> fold_optiona (fold_cExpression g) ex1_opt
                                     |> fold_optiona (fold_cExpression g) ex2_opt
                                     |> fold_cStatement g stmt
                                     |> g "CFor0" a
   | fold_cStatement g (CGoto0(id: ident, a)) st = st |>  fold_ident a g id
                                                      |> g "CGoto0" a
   | fold_cStatement g (CGotoPtr0(ex1:'a cExpression, a)) st = 
                                  st |> fold_cExpression g ex1 |> g "CGotoPtr0" a
   | fold_cStatement g (CCont0 a) st  = st |> g "CCont0" a
   | fold_cStatement g (CBreak0 a) st = st |> g "CBreak0" a
   | fold_cStatement g (CReturn0 (ex:'a cExpression optiona,a)) st = 
                                  st |> fold_optiona (fold_cExpression g) ex |> g "CReturn0" a
   | fold_cStatement g (CAsm0(_: 'a cAssemblyStatement, a)) st = 
                                  st |> g "CAsm0" a

and fold_cExpression g (CComma0 (eL:'a cExpression list, a)) st = 
                                 st |> fold(fold_cExpression g) eL |> g "CComma0" a
  | fold_cExpression g (CAssign0(aop:cAssignOp, 
                                 ex1:'a cExpression,
                                 ex2:'a cExpression,a)) st = 
                                  st |> fold_cExpression g ex1
                                     |> fold_cExpression g ex2 
                                     |> g ("CAssign0 "^toString_cAssignOp aop) a
  | fold_cExpression g (CCond0(  ex1:'a cExpression, 
                                 ex2opt: 'a cExpression optiona, (* bescheuert ! Wieso option ?*) 
                                 ex3: 'a cExpression,a)) st = 
                                  st |> fold_cExpression g ex1 
                                     |> fold_optiona (fold_cExpression g) ex2opt
                                     |> fold_cExpression g ex3 |> g "CCond0" a
  | fold_cExpression g (CBinary0(bop: cBinaryOp, ex1: 'a cExpression,ex2: 'a cExpression, a)) st =
                                 st |> fold_cExpression g ex1 
                                    |> fold_cExpression g ex2 
                                    |> g ("CBinary0 "^toString_cBinaryOp bop) a 
  | fold_cExpression g (CCast0(decl:'a cDeclaration, ex: 'a cExpression, a)) st = 
                                 st |> fold_cExpression g ex 
                                    |> fold_cDeclaration g decl
                                    |> g "CCast0" a
  | fold_cExpression g (CUnary0(unop:cUnaryOp, ex: 'a cExpression, a)) st = 
                                 st |> fold_cExpression g ex 
                                    |> g ("CUnary0 "^toString_cUnaryOp unop) a
  | fold_cExpression g (CSizeofExpr0(ex:'a cExpression, a)) st = 
                                 st |> fold_cExpression g ex    |> g "CSizeofExpr0" a
  | fold_cExpression g (CSizeofType0(decl:'a cDeclaration,a)) st = 
                                 st |> fold_cDeclaration g decl |> g "CSizeofType0" a
  | fold_cExpression g (CAlignofExpr0(ex:'a cExpression, a)) st = 
                                 st |> fold_cExpression g ex    |> g "CAlignofExpr0" a
  | fold_cExpression g (CAlignofType0(decl:'a cDeclaration, a)) st = 
                                 st |> fold_cDeclaration g decl |> g "CAlignofType0" a
  | fold_cExpression g (CComplexReal0(ex:'a cExpression, a)) st = 
                                 st |> fold_cExpression g ex |> g "CComplexReal0" a
  | fold_cExpression g (CComplexImag0(ex:'a cExpression, a)) st = 
                                 st |> fold_cExpression g ex |> g "CComplexImag0" a
  | fold_cExpression g (CIndex0(ex1:'a cExpression, ex2: 'a cExpression, a)) st =
                                 st |> fold_cExpression g ex1 
                                    |> fold_cExpression g ex2 
                                    |> g "CIndex0" a
  | fold_cExpression g (CCall0(ex:'a cExpression, argS: 'a cExpression list, a)) st =
                                 st |> fold_cExpression g ex 
                                    |> fold (fold_cExpression g) argS 
                                    |> g "CCall0" a
  | fold_cExpression g (CMember0(ex:'a cExpression, id:ident, b, a)) st =
                                 st |> fold_cExpression g ex 
                                    |> fold_ident a g id
                                    |> g ("CMember0"^((b ? (K "F")) "T")) a 
  | fold_cExpression g (CVar0(id:ident,a)) st = st |> fold_ident a g id |> g "CVar0" a
  | fold_cExpression g (CConst0(cc:'a cConstant)) st  = st |> fold_cConstant g cc
  | fold_cExpression g (CCompoundLit0(decl:'a cDeclaration,
                                      eqn: ('a cPartDesignator list * 'a cInitializer) list, a)) st =
                                 st  |> fold(fn(S,init) => 
                                             fn st => st |> fold(fold_cPartDesignator g) S
                                                         |> fold_cInitializer g init) eqn 
                                     |> fold_cDeclaration g decl 
                                     |> g "CCompoundLit0" a 
  | fold_cExpression g (CGenericSelection0(ex:'a cExpression, 
                                           eqn: ('a cDeclaration optiona*'a cExpression)list,a)) st =
                                 st |> fold_cExpression g ex 
                                    |> fold (fn (d,ex) => 
                                             fn st => st |> fold_optiona (fold_cDeclaration g) d  
                                                         |> fold_cExpression g ex) eqn  
                                    |> g "CGenericSelection0" a  
  | fold_cExpression g (CStatExpr0(stmt: 'a cStatement,a)) st =  
                                 st |> fold_cStatement g stmt |> g "CStatExpr0" a
  | fold_cExpression g (CLabAddrExpr0(id:ident,a)) st = 
                                 st |> fold_ident a g id |> g "CLabAddrExpr0" a 
  | fold_cExpression g (CBuiltinExpr0(X: 'a cBuiltinThing)) st = st |> fold_cBuiltinThing g X
  
and fold_cDeclaration g (CDecl0(_ : 'a cDeclarationSpecifier list, (* TODO *)
                                mkS: (('a cDeclarator optiona*'a cInitializer optiona) * 'a cExpression optiona) list,
                                a)) st = 
                                 st |> fold(fn ((d_o, init_o),ex_opt) =>
                                            fn st => st |> fold_optiona(fold_cInitializer g) init_o
                                                        |> fold_optiona(fold_cExpression g) ex_opt) mkS
                                    |> g "CDecl0" a
  | fold_cDeclaration g (CStaticAssert0(ex:'a cExpression, slit: 'a cStringLiteral, a)) st = 
                                 st |> fold_cExpression g ex 
                                    |> fold_cStringLiteral g slit
                                    |> g "CStaticAssert0" a

and fold_cBuiltinThing g (CBuiltinVaArg0(ex:'a cExpression,decl: 'a cDeclaration,a)) st = 
                                 st |> fold_cExpression g ex 
                                    |> fold_cDeclaration g decl
                                    |> g "CBuiltinVaArg0" a 
  | fold_cBuiltinThing g (CBuiltinOffsetOf0(d: 'a cDeclaration, _: 'a cPartDesignator list, a)) st = 
                                 st |> fold_cDeclaration g d 
                                    |> g "CBuiltinOffsetOf0" a 
  | fold_cBuiltinThing g (CBuiltinTypesCompatible0 (d1: 'a cDeclaration, d2: 'a cDeclaration,a)) st= 
                                 st |> fold_cDeclaration g d1
                                    |> fold_cDeclaration g d2 
                                    |> g "CBuiltinTypesCompatible0" a 

and  fold_cInitializer g (CInitExpr0(ex: 'a cExpression, a)) st = 
                                 st |> fold_cExpression g ex |> g "CInitExpr0" a 
   | fold_cInitializer g (CInitList0 (mms: ('a cPartDesignator list * 'a cInitializer) list,a)) st = 
                                 st |> fold(fn (a,b) => 
                                            fn st => st|> fold(fold_cPartDesignator g) a 
                                                       |> fold_cInitializer g b) mms
                                    |> g "CInitList0" a

and  fold_cPartDesignator g (CArrDesig0(ex: 'a cExpression, a)) st = 
                                 st |> fold_cExpression g ex |> g "CArrDesig0" a 
   | fold_cPartDesignator g (CMemberDesig0(id: ident, a)) st = 
                                 st |> fold_ident a g id |> g "CMemberDesig0" a 
   | fold_cPartDesignator g (CRangeDesig0(ex1: 'a cExpression, ex2: 'a cExpression, a)) st = 
                                 st |> fold_cExpression g ex1
                                    |> fold_cExpression g ex2
                                    |> g "CRangeDesig0" a 

and fold_cAttribute g (CAttr0(id: ident, exS: 'a cExpression list, a)) st =
                                 st |> fold_ident a g id
                                    |> fold(fold_cExpression g) exS 
                                    |> g "CAttr0" a 

and fold_cEnumeration g (CEnum0 (ident_opt: ident optiona,
                              exS_opt: ((ident * 'a cExpression optiona) list) optiona,
                              attrS: 'a cAttribute list, a)) st = 
                                  st |> fold_optiona(fold_ident a g) ident_opt
                                     |> fold_optiona(fold(
                                            fn (id,ex_o) =>
                                            fn st => st |> fold_ident a g id
                                                        |> fold_optiona (fold_cExpression g) ex_o)) 
                                                     exS_opt   
                                     |> fold(fold_cAttribute g) attrS
                                     |> g "CEnum0" a 


end

end (*struct *)
\<close>



end
